use std::{fs, path::Path, time::Duration};

use url::Url;

use matrix_sdk::{
	self,
	events::{
		room::{
			member::MemberEventContent,
			message::{MessageEventContent, TextMessageEventContent},
		},
		MessageEventStub, StrippedStateEventStub,
	},
	identifiers::UserId,
	Client, ClientConfig, EventEmitter, RoomState, SyncRoom, SyncSettings,
};

use phf::phf_map;

use async_trait::async_trait;
use db::DataStore;

mod config;
mod db;
mod relation;

const REACTION_MAP: phf::Map<&'static str, &'static str> = phf_map! {
	"cat" => "🐈",
	"spook" => "👀",
	"glasses" => "👓"
};

struct EventCallback {
	client: Client,
	user_id: UserId,
	statsd_client: statsd::Client,
}

impl EventCallback {
	pub fn new(client: Client, user_id: UserId, statsd_client: statsd::Client) -> Self {
		Self {
			client,
			user_id,
			statsd_client,
		}
	}
}

#[async_trait]
impl EventEmitter for EventCallback {
	//Handles room messages
	async fn on_room_message(&self, room: SyncRoom, event: &MessageEventStub<MessageEventContent>) {
		if let SyncRoom::Joined(room) = room {
			if let MessageEventStub {
				content: MessageEventContent::Text(TextMessageEventContent { body: msg_body, .. }),
				event_id,
				..
			} = event
			{
				for (k, v) in REACTION_MAP.entries() {
					if msg_body.to_lowercase().contains(k) {
						// key in message. React with value

						match relation::react(
							&self.client,
							event_id.clone(),
							room.read().await.room_id.clone(),
							v.to_string(),
						)
						.await
						{
							Ok(_) => self.statsd_client.incr(&format!("reacted.{}s", k)),
							Err(x) => {
								println!("Error while reacting: {}", x);
								self.statsd_client.incr("reacted.error");
							}
						}
					}
				}
			}
		}
	}

	//Handles incoming invitations
	async fn on_stripped_state_member(
		&self,
		sync_room: SyncRoom,
		room_member: &StrippedStateEventStub<MemberEventContent>,
		_: Option<MemberEventContent>,
	) {
		if room_member.state_key == self.user_id {
			if let RoomState::Invited(x) = sync_room {
				let room = x.read().await;
				println!("Invitation received from {}", room.display_name());
				match self.client.join_room_by_id(&room.room_id).await {
					Ok(_) => {
						println!("Joined {}", room.display_name());
						self.statsd_client.incr("joined.success");
					}
					Err(_) => {
						println!("There was an error while joining {}", room.display_name());
						self.statsd_client.incr("joined.error");
					}
				};
			}
		}
	}
}

async fn login(
	homeserver_url: String,
	username: String,
	password: String,
	config: ClientConfig,
	statsd_client: statsd::Client,
	datastore: DataStore,
) -> Result<(), matrix_sdk::Error> {
	let homeserver_url = Url::parse(&homeserver_url).expect("Couldn't parse the homeserver URL");
	let mut client = Client::new_with_config(homeserver_url, config).unwrap();

	let login_response = client
		.login(
			username,
			password,
			datastore.get_device_id(),
			Some("Cat Disruptor".to_string()),
		)
		.await?;

	println!("Logged in as {}", login_response.user_id);
	println!("Device id: {}", login_response.device_id);
	datastore.save_device_id(login_response.device_id);
	let event_callback = EventCallback::new(client.clone(), login_response.user_id, statsd_client);
	client.add_event_emitter(Box::new(event_callback)).await;
	client.sync_forever(SyncSettings::new(), |_| async {}).await;

	Ok(())
}

#[tokio::main]
async fn main() -> Result<(), matrix_sdk::Error> {
	tracing_subscriber::fmt()
		.with_env_filter(tracing_subscriber::EnvFilter::from_default_env())
		.init();
	let config = config::get_config();

	if config.homeserver_url == "changeme"
		&& config.username == "changeme"
		&& config.password == "changeme"
	{
		panic!("Please update the values in the config file!");
	}

	let statsd_client = statsd::Client::new(config.statsd_server, "cat-disruptor").unwrap();

	let store_path = fs::canonicalize(Path::new(&config.store_path)).unwrap();
	if !store_path.exists() {
		fs::create_dir(config.store_path.clone()).unwrap();
	}

	let matrix_config = ClientConfig::new()
		.store_path(store_path.clone())
		.timeout(Duration::from_secs(60)); // Auto-reconnect if we lose connection

	// We need to store the device ID so that restarting the bot doesn't create new sessions.
	// In the future we will probably store other things in the database.
	// Sqlite would be overkill otherwise.
	let datastore = DataStore::open(store_path.join("cat_disruptor.sqlite"));

	login(
		config.homeserver_url,
		config.username,
		config.password,
		matrix_config,
		statsd_client,
		datastore,
	)
	.await
}
